/** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* BiDiPro/24RF08 v0.1.0                                            24rf08.c
* Copyright (c) Oleg O. Chukaev, 2008  <oleg.chukaev@mail.ru>
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <alloc.h>
#include <dir.h>
#include <stdarg.h>

#include "defines.h"
#include "debug.h"
#include "hwctrl.h"
#include "cmdline.h"
#include "config.h"
#include "bidi.h"


#pragma warn -par


/*=======================================================================
  ⨯ 㭪権
=======================================================================*/


/* Forward declarations: cmd line parser handlers */
static int opt_special_handler (char *cmd, char *param, int extra);
static int opt_outfile_handler (char *cmd, char *param, int extra);
static int opt_debug_handler (char *cmd, char *param, int extra);
static int opt_verify_handler (char *cmd, char *param, int extra);
static int opt_read_handler (char *cmd, char *param, int extra);
static int opt_write_handler (char *cmd, char *param, int extra);
static int opt_voltage_handler (char *cmd, char *param, int extra);
static int opt_count_handler (char *cmd, char *param, int extra);
static int opt_delay_handler (char *cmd, char *param, int extra);
static int opt_port_handler (char *cmd, char *param, int extra);


/*=======================================================================
   ६
=======================================================================*/


byte eep_buff[EEPROM_SIZE];	/*   EEPROM */
byte app_buff[APP_SIZE];	/*   APP */
char my_path[MAXPATH];		/*   24rf08.exe */

char *file_eep = NULL;		/*  䠩  EEPROM */
char *file_app = NULL;		/*  䠩  APP */

int app_count = 16;		/* 쪮  뢠  APP */

int special_options = 0;	/* Help, license, version */
int cmd_read = 0;		/*  */
int cmd_write = 0;		/* 訢 */

long write_delay = 10000;	/* ६   ,  */

/* command line options */
static opt_option options[] = {
	{'V', "version", 0, opt_special_handler, SPECIAL_SHOW_VERSION,
		"show version text", NULL},
	{'L', "license", 0, opt_special_handler, SPECIAL_SHOW_LICENSE,
		"show license text", NULL},
	{'h', "help", 0, opt_special_handler, SPECIAL_SHOW_HELP,
		"show help text", NULL},
	{'e', "eeprom", 1, opt_outfile_handler, 0,
		"name of EEPROM input/output file", "file"},
	{'a', "app", 1, opt_outfile_handler, 1,
		"name of APP input/output file", "file"},
	{'r', "read", 0, opt_read_handler, 0,
		"read chip", NULL},
	{'w', "write", 0, opt_write_handler, 0,
		"write chip", NULL},
	{'c', "count", 1, opt_count_handler, 0,
		"number of bytes to write to APP", "n"},
	{'d', "debug", 0, opt_debug_handler, 0,
		"enable debug output", NULL},
	{0, "vcc", 1, opt_voltage_handler, 0,
		"set Vcc (in 100 mV units)", "voltage"},
	{0, "delay", 1, opt_delay_handler, 0,
		"set write delay (in 1 us units)", "delay"},
	{'p', "port", 1, opt_port_handler, 0,
		"set port number", "port"},
};

static const char hdr1[] =
	"BiDiPro/24RF08 v0.1.0 - AT24RF08 programming tool for BiDiPro";

static const char hdr2[] =
	"Copyright (c) 2008  Oleg O. Chukaev, <oleg.chukaev@mail.ru>";

static const char hdr3[] =
	"This program comes with ABSOLUTELY NO WARRANTY; for details type `24rf08 -L'.";

/* version message */
static const char *version_msg[] = {
	"",
	hdr1,
	hdr2,
	hdr3,
	"",
	"This software is based on:",
	"  - BiDiPro hardware driver v1.2",
	"    (C) Eduard Panchenko, Alex Plusov / CnC Lab, Andrej Arnold, 2003",
	"  - Command line parser for YASM",
	"    (C) 2001  Stanislav Karchebny <berk@madfire.net>",
	"",
	"See `24rf08.txt' or output of `24rf08 -h' for the list of options.",
	""
};

/* license message */
static const char *license_msg[] = {
	hdr1,
	hdr2,
	"",
	"This program is free software; you can redistribute it and/or modify",
	"it under the terms of the GNU General Public License as published by",
	"the Free Software Foundation; either version 2 of the License, or",
	"(at your option) any later version.",
	"",
	"This program is distributed in the hope that it will be useful,",
	"but WITHOUT ANY WARRANTY; without even the implied warranty of",
	"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the",
	"GNU General Public License for more details.",
	"",
	"You should have received a copy of the GNU General Public License",
	"along with this program; if not, write to the Free Software",
	"Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA",
	"",
};


/* help messages */
static const char *help_head =
	"usage: 24rf08 [option]* [eeprom_file.bin] [app_file.app]\n"
	"Options:\n"
;

static const char *help_tail =
	"\n"
	"Sample invocation:\n"
	"    24rf08 -r -a app.bin eeprom.bin\n"
;


/*=======================================================================
  㭪樨
=======================================================================*/


/*-----------------------------------------------------------------------
  -樨  ࠡ  宩
-----------------------------------------------------------------------*/

/* 뫠   EEPROM */
int send_eep_adr (dword adr) {
	if (write_i2c_byte (I2C_EEP | (adr >> 7) & 0x06))
		return (1);
	if (write_i2c_byte (adr & 0xff))
		return (1);
	return (0);
}


/* 뫠   APP */
int send_app_adr (dword adr) {
	if (write_i2c_byte (I2C_APP))
		return (1);
	if (write_i2c_byte (adr & 0x1f))
		return (1);
	return (0);
}


/* ⠥ EEPROM */
int read_eeprom (void) {

	dword i, addr;
	int err = 0;

	power_on ();
	sda_out ();
	for (addr = 0; addr < EEPROM_SIZE; addr += PAGE_SIZE) {
		i2c_start ();
		if (send_eep_adr (addr)) {
			err = 1; break;
		}
		i2c_start ();
		if (write_i2c_byte (I2C_EEP | 1)) {
			err = 1; break;
		}
		for (i = 0; i < PAGE_SIZE; i++) {
			if (i < PAGE_SIZE - 1) {
				eep_buff[addr+i] = read_i2c_byte (1);
			} else {
				eep_buff[addr+i] = read_i2c_byte (0);
			}
		}
		i2c_stop ();
	}
	if (err) {
		i2c_stop ();
	}
	power_off ();
	return (err);
}


/* ⠥ APP */
int read_app (void) {

	dword addr;
	int err = 0;

	power_on ();
	sda_out ();
	for (addr = 0; addr < APP_SIZE; addr++) {
		i2c_start ();
		if (send_app_adr (addr)) {
			err = 1; break;
		}
		i2c_start ();
		if (write_i2c_byte (I2C_APP | 1)) {
			err = 1; break;
		}
		app_buff[addr] = read_i2c_byte (0);
		i2c_stop ();
	}
	if (err) {
		i2c_stop ();
	}
	power_off ();
	return (err);
}


/* 뢠 EEPROM */
int write_eeprom (void) {

	dword i;
	dword adr = 0;
	int r;

	power_on ();
	sda_out ();
	while (adr < EEPROM_SIZE) {
		i2c_start ();
		if (send_eep_adr (adr)) {
			goto error;
		}
		for (i = 0; i < 16; i++) {
			if (write_i2c_byte (eep_buff[adr])) {
				goto error;
			}
			adr++;
		}
		sda (0);
		scl (1);
		sda (1);
		delay_us (write_delay);
	}
	power_off ();
	return (0);

error:
	i2c_stop ();
	power_off ();
	return (1);
}


/* 뢠 APP */
int write_app (void) {

	dword i, addr;
	int r, err = 0;

	power_on ();
	sda_out ();
	for (addr = 0; addr < app_count; addr++) {
		i2c_start ();
		if (send_app_adr (addr)) {
			err = 1; break;
		}
		if (write_i2c_byte (app_buff[addr])) {
			err = 1; break;
		}
		sda (0);
		scl (1);
		sda (1);
		delay_us (write_delay);
	}
	if (err) {
		i2c_stop ();
	}
	power_off ();
	return (err);
}


/*-----------------------------------------------------------------------
  ࠡ稪 ࠬ஢ . ப
-----------------------------------------------------------------------*/

/* 뢮 ᮮ饭  訡  . ப */
static void print_error (const char *fmt, ...) {

	va_list va;

	va_start (va, fmt);
	vfprintf (stderr, fmt, va);
	va_end (va);
	fputc ('\n', stderr);
}


/* < 䠩> */
int not_an_option_handler (char *param) {

	char s[200];

	strcpy (s, param);
	strlwr (s);

	if (strstr (s, ".bin")) {
		if (file_eep) {
			debug_printf (0, "warning: `%s' rejected: you should specify"
				" only one .bin file\n", param);
		} else {
			file_eep = strdup (param);
		}
	} else if (strstr (s, ".app")) {
		if (file_app) {
			debug_printf (0, "warning: `%s' rejected: you should specify"
				" only one .app file\n", param);
		} else {
			file_app = strdup (param);
		}
	} else {
		debug_printf (0, "warning: `%s' rejected: please use .bin and"
			" .app extensions\n", param);
	}
	return (0);
}


/* <騩 > */
int other_option_handler (char *option) {
	return 1;
}


/* -h, -L, -V */
static int opt_special_handler (char *cmd, char *param, int extra) {

	if (special_options == 0)
		special_options = extra;
	return 0;
}


/* -e, -a */
static int opt_outfile_handler (char *cmd, char *param, int extra) {

	if (extra == 0) {
		if (file_eep) {
			debug_printf (0, "warning: `%s' rejected: you should specify"
				" only one .bin file\n", param);
		} else {
			file_eep = strdup (param);
		}
	} else if (extra == 1) {
		if (file_app) {
			debug_printf (0, "warning: `%s' rejected: you should specify"
				" only one .app file\n", param);
		} else {
			file_app = strdup (param);
		}
	}
	return (0);
}


/* -d, -Q */
static int opt_debug_handler (char *cmd, char *param, int extra) {

       	debug_level++;
	return (0);
}


/* -r */
static int opt_read_handler (char *cmd, char *param, int extra) {

       	cmd_read = 1;
	return (0);
}


/* -w */
static int opt_write_handler (char *cmd, char *param, int extra) {

       	cmd_write = 1;
	return (0);
}


/* --vcc */
static int opt_voltage_handler (char *cmd, char *param, int extra) {

	long i;

	i = atol (param);
	if (i <= 0)
		return (1);
	vcc = i;
	return (0);
}


/* -c */
static int opt_count_handler (char *cmd, char *param, int extra) {

	long i;

	i = atol (param);
	if (i <= 0)
		return (1);
	if (i > APP_SIZE)
		i = APP_SIZE;
	app_count = i;
	return (0);
}


/* --delay */
static int opt_delay_handler (char *cmd, char *param, int extra) {

	long i;

	i = atol (param);
	if (i == 0)
		return (1);
	write_delay = i;
	return (0);
}


/* -p */
static int opt_port_handler (char *cmd, char *param, int extra) {

	port_number = atol (param);
	return (0);
}


/*-----------------------------------------------------------------------
  Main
-----------------------------------------------------------------------*/

int main (int argc, char *argv[]) {

	FILE *f;
	int rc, cmd;
	size_t i;
	char *cp, *p, *cmdlst[20], cmdbuff[1000], line[200];

	strcpy (my_path, argv[0]);
	p = strrchr (my_path, '\\');
	*++p = '\0';
	strcpy (progname, "24rf08");

	/* Import ini file as command line parameters */
	cmd = 1; cp = cmdbuff+1; cmdbuff[0] = '\0'; cmdlst[0] = cmdbuff;
	f = open_cfgfile (INIFILE);
	if (f != NULL) {
		while (fgets (line, sizeof(line), f) != NULL) {
			if (line[0] != ';' && line[0] != '\n') {
				if (cmd >= NELEMS(cmdlst) - 1)
					break;
				strncpy (cp, line, cmdbuff + sizeof(cmdbuff) - cp);
				cmdlst[cmd++] = cp;
				p = strchr (cp, '\n');
				if (p) *p = '\0';
				cp += strlen (cp) + 1;
			}
		}
		fclose (f);
	}

	if (parse_cmdline (cmd, cmdlst, options, NELEMS(options), print_error))
		return (1);

	if (parse_cmdline (argc, argv, options, NELEMS(options), print_error))
		return (1);

	switch (special_options) {
	case SPECIAL_SHOW_HELP:
		/* -h, --help */
		printf ("%s\n%s\n\n", hdr1, hdr2);
		help_msg (help_head, help_tail, options, NELEMS(options));
		return (0);
	case SPECIAL_SHOW_VERSION:
		/* -V, --version */
		for (i = 0; i < NELEMS(version_msg); i++)
			printf ("%s\n", version_msg[i]);
		return (0);
	case SPECIAL_SHOW_LICENSE:
		/* -L, --license */
		for (i = 0; i < NELEMS(license_msg); i++)
			printf("%s\n", license_msg[i]);
		return (0);
	}

	if (cmd_read == 0 && cmd_write == 0 &&
		file_eep == NULL && file_app == NULL)
	{
		for (i = 0; i < NELEMS(version_msg); i++)
			printf ("%s\n", version_msg[i]);
		return (0);
	}

	if (cmd_read == 0 && cmd_write == 0) {
		debug_printf (0, "error: -r or -w option must be specified\n");
		return (1);
	}

	if (file_eep == NULL && file_app == NULL) {
		debug_printf (0, "error: at least one file must be specified\n");
		return (1);
	}

	rc = initialize_port ();
	if (rc) {
		debug_printf (0, "error: can't initialize hardware: 0x%04X\n", rc);
		return (1);
	}

	if (cmd_read) {
		/* APP */
		if (file_app != NULL) {
			debug_printf (1, "Reading APP...\n");
			if (read_app ()) {
			       	debug_printf (0, "error: can't read APP\n");
				return (1);
			}
			f = fopen (file_app, "wb");
			if (f == NULL) {
				debug_printf (0, "error: can't open `%s'\n",
					file_app);
				return (1);
			}
			fwrite (app_buff, APP_SIZE, 1, f);
			fclose (f);
		}
		/* EEPROM */
		if (file_eep != NULL) {
			debug_printf (1, "Reading EEPROM...\n");
			if (read_eeprom ()) {
			       	debug_printf (0, "error: can't read EEPROM\n");
				return (1);
			}
			f = fopen (file_eep, "wb");
			if (f == NULL) {
				debug_printf (0, "error: can't open `%s'\n",
					file_eep);
				return (1);
			}
			fwrite (eep_buff, EEPROM_SIZE, 1, f);
			fclose (f);
		}
	} else if (cmd_write) {
		/* APP */
		if (file_app != NULL) {
			f = fopen (file_app, "rb");
			if (f == NULL) {
				debug_printf (0, "error: can't open `%s'\n",
					file_app);
				return (1);
			}
			fread (app_buff, 1, APP_SIZE, f);
			fclose (f);
			debug_printf (1, "Writing APP (%d bytes)...\n", app_count);
			if (write_app ()) {
			       	debug_printf (0, "error: can't write APP\n");
				return (1);
			}
		}
		/* EEPROM */
		if (file_eep != NULL) {
			f = fopen (file_eep, "rb");
			if (f == NULL) {
				debug_printf (0, "error: can't open `%s'\n",
					file_eep);
				return (1);
			}
			fread (eep_buff, EEPROM_SIZE, 1, f);
			fclose (f);
			debug_printf (1, "Writing EEPROM...\n");
			if (write_eeprom ()) {
			       	debug_printf (0, "error: can't write EEPROM\n");
				return (1);
			}
		}
	}

	return (0);
}

/** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* E0F
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **/

