// Title:  BiDiPro DLL module
// Desc:   16-bit Atmel Flashes AT49Fxxx
// Ver:    0.0.1
// Found:  2007-08-22
// Last:   2007-08-22
// CPU:    386+
// Model:  Tiny
// Compil: Borland C++ 3.1 (DOS)
// Author: Oleg O. Chukaev

// Based on Atmel datasheets


#include <dos.h>
#include <conio.h>
#include <stdlib.h>
#include "..\dlldata.h"
#include "..\errors.h"
#include "..\bidi.h"

//#define CMDLINE

#ifndef CMDLINE
 #define dlp ((DLLDATA *) NULL)        /* ⥫    */
#else
 extern DLLDATA far * dlp;
#endif //CMDLINE

#define bdport dlp->bdprt            /* 㪠⥫  hardware ࠩ */

#define DataPortIn()  bdport(PFORCE|PWR1|PAIN|PBIN)
#define DataPortOut() bdport(PFORCE|PWR1)

#define ATMP 64


struct  CHDEFS {
	dword size;   // 
	byte  dpins;  //   dip 
	byte  rA2;    // ॣ  ⥭ //WE=1,CE=0,OE=0
	byte  rC1;    // use only for 256
	byte  rwB2;   // 殮
	byte  wC1h;   // :
	byte  wC1l;   //
	byte  wA2h;   // WE=1,CE=1,OE=1
	byte  wA2l;   // WE=0,CE=0,OE=1
};

CHDEFS chd[]={
//   size    pins  rA2   rC1  rwB2  wC1h  wC1l  wA2h  wA2l
  {0x008000L, 32, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},  //256
  {0x010000L, 32, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},  //516
  {0x020000L, 32, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},  //1024
  {0x040000L, 32, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},  //2048
  {0x080000L, 32, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},  //4096
  {0x100000L, 32, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},  //8192
};


// ७   CFG:
static struct CFGKEY cfgkeys[] = {
{'D', "Vccp",  50,50,70},      //殮 ⠭  ணࠬ஢   0
{'D', "Vccv",  50,50,70},      //殮 ⠭  䨪樨        1
{'D', "ErTime",10000,10000,0}, //p 横 p (ms)                 2
{'D', "WrTime",50,50,0},       //p  ᫮ (mcs)                  3
{'D', "Nnum",  25,25,0},       //᫮ ⮪ ணࠬ஢            4
{0,    NULL,   0,0},           //End of table
};

#define cfg_Vccp   cfgkeys[0].data
#define cfg_Vccv   cfgkeys[1].data
#define cfg_ErTime cfgkeys[2].data
#define cfg_WrTime cfgkeys[3].data
#define cfg_Nnum   cfgkeys[4].data
//#define cfg_Page  cfgkeys[5].data

static char *copyright = "16-bit Atmel Flash module (c) Oleg O. Chukaev ["__DATE__"]";
static word chtype = 0xFFFF;           //  ROM
static CHDEFS *pchd;

#define power_off breset


void breset (void) {
	bdport (PRST);
}


void dbgBox (char *str) {
	dlp->msgBox (dmfError | dmfOKButton, str);
}


int power_on (byte mode) {

	word vmask = 0x80;
	word vid = 140;
	/* Vid   ⠭  14V  */
	/* 祭 ⠡쭮 0  A9 */

	if (chtype == 0xFFFF)
		return (ERC_CHIPUNSP);

	if (mode == 'R') {
		DataPortIn ();			/* DD1:IIO */
		bdport (PFORCE | PWR2);		/* DD5:OOO */
		bdport (PFORCE | PWA2 | BITTER | BITA18 | BITA19);
		dlp->voltage (dlp->Vcc, dlp->Vpp, vid);
		bdport (PWB2 | vmask);
	} else if (mode == 'W') {
		bdport (PFORCE | PWR2);		/* DD5:OOO */
		bdport (PFORCE | PWA2 | BITTER | BITA18 | BITA19);
		DataPortOut ();			/* DD1:OOO */
		dlp->voltage (dlp->Vcc, dlp->Vpp, vid);
		bdport (PWB2 | vmask );		/* Vcc on */
	} else if (mode == 'I') {
		bdport (PFORCE | PWR2);
		bdport (PFORCE | PWA2 | BITTER | BITA18 | BITA19);
		DataPortOut ();
        	vid = dlp->Vid;
		dlp->voltage (dlp->Vcc, dlp->Vpp, vid);
		if (dlp->Vid > 0)
			vmask |= 0x10;		/* Vid enable */
		bdport (PWB2 | vmask );		/* Vcc, Vid on */
	}
	dlp->mcsdelay (100000L);
	return (0);
}


void write_code16 (word addr, word data) {
	bdport (PWIR | (byte)((addr >> 0) & 0xff));	/* A0..A7 */
	bdport (PWC1 | (byte)((addr >> 8) & 0xff));	/* A8..A15 */
	bdport (PWA2 | BITTER | BITOE | BITA18 | BITA19
		| (byte)((addr >> 16) & 0x03));		/* A16..A17 */
        bdport (PWA1 | (data & 0xff));
        bdport (PWB1 | (data >> 8));
	bdport (PWA2 | BITTER | BITOE | BITA19 | (byte)((addr >> 16) & 0x03));
	dlp->mcsdelay (2);
	bdport (PWA2 | BITTER | BITOE | BITA18 | BITA19
		| (byte)((addr >> 16) & 0x03));
}


/* ⥭ ᫮. adr  ᮬ _᫮_ */
inline word read_one (dword adr) {

	/* A0..A7 */
	bdport (PWIR | (byte)((adr >> 0) & 0xff));
	/* A8..A15 */
	bdport (PWC1 | (byte)((adr >> 8) & 0xff));
        /* A16..A17, -CE == 0, -OE == 0, -WE == 1, -RESET == 1 */
	bdport (PWA2 | BITTER | BITA18 | BITA19 | (byte)((adr >> 16) & 0x03));
        /* ⠥  */
	return (((byte)bdport(PRB1) << 8) + (byte)bdport(PRA1));
}


/*  ᫮. adr  ᮬ _᫮_ */
void write_one (dword adr, word dat) {
	write_code16 (0x5555, 0xaa);
	write_code16 (0x2aaa, 0x55);
	write_code16 (0x5555, 0xa0);
	write_code16 (adr, dat);
}


/* ⥭ ᥣ 稯. */
int read_all (void) {

	dword adr;
	dword istep = pchd->size / 64L;
        word dat;

	if (chtype == 0xFFFF)
		return (ERC_CHIPUNSP);

	power_on ('R');
	for (adr = 0; adr < pchd->size; adr += 2) {
        	dat = read_one (adr >> 1);
		dlp->setbytebuf (adr, dat & 0xff);
		dlp->setbytebuf (adr + 1, dat >> 8);
		if (adr % istep == 0 && dlp->iprogress (adr >> 1, 2))
			break;
	}
	power_off ();
	return (0);
}


/*  ᥣ 稯. */
int write_all (void) {

	dword adr;
	word n, t;
	dword istep = pchd->size / 64L;
	word tdr, tdw;

	if (chtype == 0xFFFF)
		return (ERC_CHIPUNSP);

	power_on ('W');
	for (adr = 0; adr < pchd->size; adr += 2) {
		tdw = dlp->getbytebuf (adr) + (dlp->getbytebuf (adr+1) << 8);
		for (n = 0; n <= cfg_Nnum; n++) {
			DataPortOut ();
                        write_one (adr >> 1, tdw);

			DataPortIn ();
			for (t = 0; t < cfg_WrTime; t++) {
				dlp->mcsdelay (1);
				tdr = read_one (adr >> 1);
				if ((tdr & 0x80) == (tdw & 0x80))
					break; //right data
			}

			if (tdr == tdw)
				break;
			if (n == cfg_Nnum) {   // error
				dlp->adr = adr;
				dlp->edata = tdr;
				power_off ();
				return (ERC_WRITE);
			}
		}
		if (adr % istep == 0 && dlp->iprogress (adr >> 1, 2))
			break;
	}
	power_off ();
	return (0);
}


/* ࠭ 稯. */
int erase (void) {

	dword adr, j;
	word tdr = 0;
	int i, er_ok = 0;

	if (chtype == 0xFFFF)
		return (ERC_CHIPUNSP);

	dlp->iprogress (ATMP, 0);
	power_on ('W');

	for (i = 1; i <= ATMP; i++) {
		DataPortOut ();             // DD1:OIO
                write_code16 (0x5555, 0xaa);
                write_code16 (0x2aaa, 0x55);
                write_code16 (0x5555, 0x80);
                write_code16 (0x5555, 0xaa);
                write_code16 (0x2aaa, 0x55);
                write_code16 (0x5555, 0x10);

		if (dlp->iprogress (i, 2))
			break;  //progress

		DataPortIn ();
		bdport (PWA2 | BITTER | BITA18 | BITA19);
                tdr = read_one (0) & 0x40;
		bdport (PWA2 | BITTER | BITA18 | BITA19 | BITOE);
                for (j = 0; j < cfg_ErTime; j++) {
                	dlp->mcsdelay (1000);
			bdport (PWA2 | BITTER | BITA18 | BITA19);
			if ((read_one (0) & 0x40) == tdr) {
                        	er_ok = 1;
				break;
                        } else {
                        	tdr ^= 0x40;
                        }
			bdport (PWA2 | BITTER | BITA18 | BITA19 | BITOE);
                }
		if (er_ok)
                	break;
	}
	dlp->iprogress (0, 1); //close
	power_off ();

	return 0;
}


/* ⥭ 䨪. */
int read_id (void) {

	dword id1, id2;
	dlp->edata = 0;

	if (chtype == 0xFFFF)
		return (ERC_CHIPUNSP);

	power_on ('I');

	if (dlp->Vid < 120) {
	        /* Intelligent (software) ID */
		DataPortOut ();
		write_code16 (0x5555, 0xaa);
		write_code16 (0x2aaa, 0x55);
		write_code16 (0x5555, 0x90);
		bdport (PWA2 | BITTER | BITA18 | BITA19);

		DataPortIn ();			/* DD1:IIO */
		bdport (PWC1 | 0);		/* 2   */
		bdport (PWIR | 0);		/* 1   */
		id1 = (byte)bdport (PRA1);	/* Mfr code */
		bdport (PWIR | 1);		/* 1   = 1 */
		id2 = (byte)bdport (PRA1);	/* Dev code */

		DataPortOut ();
		write_code16 (0, 0xf0);
	} else {
		/* Hardware ID */
		bdport (PWA2 | BITTER | BITA18 | BITA19 | BITPE);
		dlp->mcsdelay (10);

		DataPortIn ();			/* DD1:IIO */
		bdport (PWC1 | 0);		/* 2   */
		bdport (PWIR | 0);		/* 1   = 0 */
		id1 = (byte)bdport (PRA1);	/* Mfr code */
		bdport (PWIR | 1);		/* 1   = 1 */
		id2 = (byte)bdport (PRA1);	/* Dev code */

		bdport (PWA2 | BITTER | BITA18 | BITA19);
	}

	dlp->edata = id1 | (id2 << 16L);
	power_off ();
	return 0;
}


/*  祭 -. */
int test (void) {

	dword id;

	if (chtype == 0xFFFF)
		return (ERC_CHIPUNSP);

	power_on ('I');

        /* Intelligent (software) ID */
	DataPortOut ();
	write_code16 (0x5555, 0xaa);
	write_code16 (0x2aaa, 0x55);
	write_code16 (0x5555, 0x90);
	bdport (PWA2 | BITTER | BITA18 | BITA19);

	DataPortIn ();			/* DD1:IIO */
	bdport (PWC1 | 0);		/* 2   */
	bdport (PWIR | 2);		/* 1   */
	id = (byte)bdport (PRA1);	/* code */

	DataPortOut ();
	write_code16 (0, 0xf0);

	power_off ();

        if (id & 1) {
        	dlp->msgBox (dmfInformation | dmfOKButton,
			"\003Boot Block is locked");
        } else {
        	dlp->msgBox (dmfInformation | dmfOKButton,
                	"\003Boot Block is not locked");
        }

	return (0);
}


/* ࠢ 稯  ஬. */
int compare_all (void) {

	dword i;
	dword errors = 0;
	dword istep = pchd->size / 64L;
        word dat;

	if (chtype == 0xFFFF)
		return ERC_CHIPUNSP;

	power_on ('R');
	for (i = 0; i < pchd->size; i += 2) {
        	dat = read_one (i >> 1);
		if ((dlp->getbytebuf (i) + (dlp->getbytebuf (i+1) << 8)) != dat) {
                	/* ⬥ ᮢ */
			dlp->setbmark (i, 1);
			dlp->setbmark (i + 1, 1);
			errors += 2;
		}
		if (i % istep == 0 && dlp->iprogress (i >> 1, 2))
			break;
	}
	power_off ();
	dlp->edata = errors;
	return (0);
}


/* ஢ઠ  . */
int blank_chk (void) {

	dword i;
	dword istep = pchd->size / 64L;

	if (chtype == 0xFFFF)
		return (ERC_CHIPUNSP);

	power_on ('R');
	for (i = 0; i < pchd->size; i += 2) {
		if (read_one (i >> 1) != 0xffff) {
			dlp->adr = i;
			power_off ();
			return (1);
		}
		if (i % istep == 0) {
			if (dlp->iprogress (i >> 1, 2))
				break;
		}
	}
	power_off ();
	return (0);
}


/* ⠭ ⨯ ROM */
word settype (void) {

        word i;

	for (i = 0; i < sizeof(chd) / sizeof(CHDEFS); i++) {
		if (chd[i].size == dlp->esize)
			return (chtype = i);
	}
	return (chtype = 0xFFFF);
}


/* main code */
extern "C" int dmain (void) {

	switch (dlp->funcode) {
	case DFN_LOAD:
		/* ⨢஢ 㯭 㭪樨: */
		dlp->funmask[DFN_FREAD] = 1;
		dlp->funmask[DFN_FWRITE] = 1;
		dlp->funmask[DFN_READID] = 1;
		dlp->funmask[DFN_ERASE] = 1;
		dlp->funmask[DFN_FCOMPARE] = 1;
		dlp->funmask[DFN_FBLANK] = 1;
		dlp->funmask[DFN_TEST] = 1;
		dlp->defstate = 0xFF;
		dlp->copyright = copyright;
		dlp->version = DLLVERSION;
		dlp->cfgkeys = cfgkeys;
		dlp->Vppmax = 140;
		return (0);

	case DFN_INIT:
		if (settype () == 0xFFFF)
			return (ERC_CHIPUNSP);
		pchd = &chd[chtype];
		dlp->dippins = pchd->dpins;
		return (0);

	case DFN_FREAD:
		return (read_all ());

	case DFN_FWRITE:
		return (write_all ());

	case DFN_FCOMPARE:
		return (compare_all ());

	case DFN_FBLANK:
		return (blank_chk ());

	case DFN_READID:
		return (read_id ());

	case DFN_ERASE:
		return (erase ());

	case DFN_TEST:
		return (test ());

	default:
		return (ERC_FUNUNSP);
	}
}


/* E0F */

